#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>	/* O_WRONLY and friends */
#include <limits.h>

#if !defined(__linux__)
#error "Platform not appropriate"
#endif

#include "linux/soundcard.h"
#include "bcm_backend_ioctl.h"

#define HAL_MAX_PLAY_OUTPUTS	2		/* numbered 0, 1 */
#define HAL_LINUX_PCMDRV_LEN	32
#define HAL_LINUX_PCMDRV	"/dev/brcmpcm_0"
#define HAL_OUTPUT_CHANNEL	1
#define HAL_OUTPUT_SAMPLE_RATE	44100

/* datatype aliases */

#define HAL_BCM_SAMPLE_RATE_CTRL bcm_sample_rate_ctrl
#define HAL_BRCM_IOCTL_SET_SAMPLE_RATE BRCM_IOCTL_SET_SAMPLE_RATE
#define HAL_BCM_AUDIO_AUTOREPEAT_CTRL bcm_audio_autorepeat_ctrl
#define HAL_BRCM_IOCTL_AUDIO_SET_AUTOREPEAT BRCM_IOCTL_AUDIO_SET_AUTOREPEAT

#define HAL_TONE_SAMPLE_MULTIPLIER (8)   /* num sine waves in buffer */
#define HAL_TONE_SAMPLE_MAX HAL_TONE_SAMPLE_MULTIPLIER*sizeof(samples)

#if defined(DEBUG)
#define	TRACE_IO(x)	fprintf(stderr, "%s\n", x);
#else
#define TRACE_IO(x)	/* use -DDEBUG for trace IO here */
#endif

/* For a Sine wave */
   /* When the hardware is silent, this is equivalent to having a
    * sample with all data at 127 - so centre the wave at this value.
    * Otherwise we get a noticeable volume spike as the tone starts
    */
//#define HAL_TONE_VOLUME ((UCHAR_MAX/2)*0.21)	/* 560mV on DP411 */
#define HAL_TONE_VOLUME ((UCHAR_MAX/2)*0.13)	/* as was - '18' */
const u_int8_t samples[]= {
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * 0),
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * 0.707),
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * 1.0),
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * 0.707),
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * 0),
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * -0.707),
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * -1.0),
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * -0.707)
		};

int16_t tone_fd[HAL_MAX_PLAY_OUTPUTS]=
{
   -1,
   -1
};

int tone_hw_open(const int chan)
{
	if (chan >= HAL_MAX_PLAY_OUTPUTS)
	{
	    return -1;
	}
	else if (tone_fd[chan] != -1)
	{
	    return -1;	/* already open */
	}
	else
	{
	  char tone_driver[HAL_LINUX_PCMDRV_LEN];

	    snprintf(tone_driver, HAL_LINUX_PCMDRV_LEN, "%s_%d", HAL_LINUX_PCMDRV, chan);
	    if ((tone_fd[chan]= open(tone_driver, O_WRONLY)) < 0)
	    {
		return -1;
	    }

	  return 0;
	}
}

int tone_hw_init(const int chan)
{
	if (chan >= HAL_MAX_PLAY_OUTPUTS)
	{
	    TRACE_IO("Bad channel")
	    return -1;
	}
	else if (tone_fd[chan] == -1)
	{
	    TRACE_IO("Channel not open")
	    return -1;	/* already open */
	}
	else
	{
	  HAL_BCM_SAMPLE_RATE_CTRL	rate;
	  HAL_BCM_AUDIO_AUTOREPEAT_CTRL	repeat;
	  u_int32_t			argument;
	  u_int32_t			speed;

	    argument= HAL_OUTPUT_CHANNEL;
	    if (ioctl(tone_fd[chan], SNDCTL_DSP_CHANNELS, &argument) == -1)
	    {
		TRACE_IO("SNDCTL_DSP_CHANNELS")
		return -1;
	    }

	    /* set unsigned sample format (volume-zero at n=128)
	     * ...we assume the hardware does not need to be told that
	     * AFMT_U8 implies eight bit samples, and we assume it does
	     * mono output by default
	     */
	    argument= (u_int32_t)AFMT_U8;
	    if (ioctl(tone_fd[chan], SNDCTL_DSP_SETFMT, &argument) == -1)
	    {
		TRACE_IO("SNDCTL_DSP_SETFMT")
		return -1;
	    }

	    /* Set 'rate.ratio' to something sane compared to the set of
	     * SNDCTL_DSP_SPEED settings we'll use to generate the
	     * different tones. This way round appears to allow us a) to
	     * keep the sample we send small, and b) only to send the
	     * sample to the hardware once.
	     */
	    rate.ratio= HAL_OUTPUT_SAMPLE_RATE;
	    if (ioctl(tone_fd[chan], HAL_BRCM_IOCTL_SET_SAMPLE_RATE, &rate) == -1)
	    {
		TRACE_IO("IOCTL_SET_SAMPLE_RATE")
		return -1;
	    }

	    repeat.enable= 1;
	    if (ioctl(tone_fd[chan], HAL_BRCM_IOCTL_AUDIO_SET_AUTOREPEAT, &repeat) == -1)
	    {
		TRACE_IO("IOCTL_AUDIO_SET_AUTOREPEAT")
		return -1;
	    }
	}

  return 0;
}

int tone_hw_send(int chan, u_int32_t speed, int secs)
{
	if (chan >= HAL_MAX_PLAY_OUTPUTS)
	{
	    TRACE_IO("Bad channel")
	    return -1;
	}
	else if (tone_fd[chan] == -1)
	{
	    TRACE_IO("Channel not open")
	    return -1;	/* already open */
	}
	else
	{
	   u_int8_t	sample_block[HAL_TONE_SAMPLE_MAX];
	   int		ix;

	    for (ix=0; ix < HAL_TONE_SAMPLE_MAX; ix++)
	    {
		sample_block[ix]= (samples[ix % sizeof(samples)]);
	    }

	    /* NB: write() traditionally returns (ssize_t)n > 0 [successful
	     * output quantity], but broadcom PCM hardware doesn't appear
	     * to do this *when 'autorepeat' is TRUE*
	     */
	    if (write(tone_fd[chan], &sample_block, HAL_TONE_SAMPLE_MAX) != 0 )
	    {
		return -1;
	    }

	    if (ioctl(tone_fd[chan], SNDCTL_DSP_SPEED, &speed) == -1)
	    {
		/* Change SNDCTL_DSP_SPEED (traditionally the output sample
		 * quality ... but see comment about only sending the sample to
		 * the hardware once).
		 */
		return -1;
	    }

	    sleep(secs);

	  return 0;
	}
}

int tone_hw_close(const int chan)
{
	if (chan >= HAL_MAX_PLAY_OUTPUTS)
	{
	    return -1;
	}
	else if (tone_fd[chan] == -1)
	{
	    return -1;
	}

	if (close(tone_fd[chan]) == -1)
	    return -1;

	tone_fd[chan]= -1;
  return 0;
}


int main(const int argc, const char **argv)
{
  int freq;
	if (argc == 0)
	{
	    freq= 4*1300;	/* 1300 is HAL_TONE_BASE */
	}
	else
	{
	    freq= atoi(argv[1]);
	    if (freq < 100)
	    {
		fprintf(stderr, "%s: Freq-arg %d is ridiculous\n", argv[0], freq);
		exit(1);
	    }
	}

	printf("Open world %d\n", tone_hw_open(0));
	printf("Init world %d\n", tone_hw_init(0));
	printf("Send world %d\n", tone_hw_send(0, freq, 1));
	printf("Close world %d\n", tone_hw_close(0));

  return 0;
}
