#define __USE_BSD	/* snprintf() in stdio.h */
#include <stdio.h>
#include <fcntl.h>	/* O_WRONLY and friends */
#include <limits.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>	/* write(), close() */
#include <sys/ioctl.h>

#if !defined(__linux__)
#error "Platform not appropriate"
#endif

#include "linux/soundcard.h"
#if !defined(PURE_OSS)
#include "bcm_backend_ioctl.h"

#define HAL_LINUX_PCMDRV_LEN	32
#define HAL_LINUX_PCMDRV	"/dev/brcmpcm_0"
#endif

#define MAX_SAMP_KB		512
#define BYTES_PER_KB		1024

#define HAL_MAX_PLAY_OUTPUTS	2		/* numbered 0, 1 */
#define HAL_OUTPUT_CHANNEL	1
#define HAL_OUTPUT_SAMPLE_RATE	44100

/* datatype aliases */

#if !defined(PURE_OSS)
#define HAL_BCM_SAMPLE_RATE_CTRL bcm_sample_rate_ctrl
#define HAL_BRCM_IOCTL_SET_SAMPLE_RATE BRCM_IOCTL_SET_SAMPLE_RATE
#define HAL_BCM_AUDIO_AUTOREPEAT_CTRL bcm_audio_autorepeat_ctrl
#define HAL_BRCM_IOCTL_AUDIO_SET_AUTOREPEAT BRCM_IOCTL_AUDIO_SET_AUTOREPEAT
#endif

#define HAL_TONE_SAMPLE_MULTIPLIER (8)   /* num sine waves in buffer */
#define HAL_TONE_SAMPLE_MAX HAL_TONE_SAMPLE_MULTIPLIER*sizeof(samples)

#if defined(DEBUG)
#define	TRACE_IO(x)	fprintf(stderr, "%s\n", x);
#else
#define TRACE_IO(x)	/* use -DDEBUG for trace IO here */
#endif

/* For a Sine wave */
   /* When the hardware is silent, this is equivalent to having a
    * sample with all data at 127 - so centre the wave at this value.
    * Otherwise we get a noticeable volume spike as the tone starts
    */
#define HAL_TONE_VOLUME ((UCHAR_MAX/2)*0.13)	/* as was - '18' */
const u_int8_t samples[]= {
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * 0),
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * 0.707),
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * 1.0),
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * 0.707),
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * 0),
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * -0.707),
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * -1.0),
		(u_int8_t)(UCHAR_MAX/2 + HAL_TONE_VOLUME * -0.707)
		};

int16_t audout_fd[HAL_MAX_PLAY_OUTPUTS]=
{
   -1,
   -1
};

/* prototypes */

void keyb_init(struct termios *sanitty);
void keyb_halt(struct termios *sanitty);

int audout_open(const int chanID);
int audout_config(const int chanID);
int audout_sine(int chanID, int speed);
int audout_close(const int chanID);

int process_sample(int chanID, int speed, FILE *samfh);

int audout_open(const int chanID)
{
	if (chanID >= HAL_MAX_PLAY_OUTPUTS)
	{
	    return -1;
	}
	else if (audout_fd[chanID] != -1)
	{
	    return -1;	/* already open */
	}
	else
	{
#if defined(PURE_OSS)
	    if ((audout_fd[chanID]= open("/dev/dsp", O_WRONLY)) == -1)
	    {
		return -1;
	    }
#else
	  char tone_driver[HAL_LINUX_PCMDRV_LEN];

	    snprintf(tone_driver, HAL_LINUX_PCMDRV_LEN, "%s_%d", HAL_LINUX_PCMDRV, chanID);
	    if ((audout_fd[chanID]= open(tone_driver, O_WRONLY)) < 0)
	    {
		return -1;
	    }
#endif

	  return 0;
	}
}


/* declarations */

void keyb_init(struct termios *sanitty)
{
  struct termios quietty;

	/* keyboard */
	if (tcgetattr(fileno(stdin), sanitty) == -1)
	{
	    perror ("tcgetattr");
	    exit (-1);
	}

	quietty.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
	quietty.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
	quietty.c_cflag &= ~(CSIZE | PARENB);
	quietty.c_cflag |= CS8;
	quietty.c_oflag &= ~(OPOST);
	quietty.c_cc[VMIN] = 1;
	quietty.c_cc[VTIME] = 1;

	if (tcsetattr(fileno(stdin), TCSAFLUSH, &quietty) == -1)
	{
	    perror ("tcgetattr");
	    exit (-1);
	}
}

void keyb_halt(struct termios *sanitty)
{
	if (tcsetattr(fileno(stdin), TCSAFLUSH, sanitty) == -1)
	{
	    perror ("tcgetattr");
	    exit (-1);
	}
}

int audout_config(const int chanID)
{
	if (chanID >= HAL_MAX_PLAY_OUTPUTS)
	{
	    TRACE_IO("Bad chanID")
	    return -1;
	}
	else if (audout_fd[chanID] == -1)
	{
	    TRACE_IO("Channel not open")
	    return -1;	/* already open */
	}
	else
	{
	  int			argument;

#if defined(SNDCTL_SETSONG)
	    if (ioctl(audout_fd[chanID], SNDCTL_SETSONG, __FILE__) == -1)
	    {
		TRACE_IO("SNDCTL_DSP_SETSONG")
		return -1;
	    }
#endif

	    argument= HAL_OUTPUT_CHANNEL;
	    if (ioctl(audout_fd[chanID], SNDCTL_DSP_CHANNELS, &argument) == -1)
	    {
		TRACE_IO("SNDCTL_DSP_CHANNELS")
		return -1;
	    }
	    if (argument != HAL_OUTPUT_CHANNEL)
	    {
		TRACE_IO("SNDCTL_DSP_CHANNELS: request not supported")
		return -1;
	    }

	    /* set unsigned sample format (volume-zero at n=128)
	     * ...we assume the hardware does not need to be told that
	     * AFMT_U8 implies eight bit samples, and we assume it does
	     * mono output by default
	     */
	    argument= (int)AFMT_U8;
	    if (ioctl(audout_fd[chanID], SNDCTL_DSP_SETFMT, &argument) == -1)
	    {
		TRACE_IO("SNDCTL_DSP_SETFMT")
		return -1;
	    }
	    if (argument != AFMT_U8)
	    {
		TRACE_IO("SNDCTL_DSP_SETFMT: AFMT_U8 not supported")
		return -1;
	    }

	    argument= (int)0;	/* FALSE -> mono */
	    if (ioctl(audout_fd[chanID], SNDCTL_DSP_STEREO, &argument) == -1)
	    {
		TRACE_IO("SNDCTL_DSP_STEREO")
		return -1;
	    }
	    if (argument != 0)
	    {
		TRACE_IO("SNDCTL_DSP_STEREO: request not supported")
		return -1;
	    }
	}

  return 0;
}

int audout_sine(int chanID, int speed)
{
	if (chanID >= HAL_MAX_PLAY_OUTPUTS)
	{
	    TRACE_IO("Bad chanID")
	    return -1;
	}
	else if (audout_fd[chanID] == -1)
	{
	    TRACE_IO("Channel not open")
	    return -1;	/* already open */
	}
	else
	{
	  u_int8_t	sample_block[HAL_TONE_SAMPLE_MAX];
	  unsigned int	ix;
#if !defined(PURE_OSS)
	  HAL_BCM_AUDIO_AUTOREPEAT_CTRL	repeat;
	  HAL_BCM_SAMPLE_RATE_CTRL	rate;
#endif
	  struct termios sanitty;

	    keyb_init(&sanitty);

	    for (ix=0; ix < HAL_TONE_SAMPLE_MAX; ix++)
	    {
		sample_block[ix]= (samples[ix % sizeof(samples)]);
	    }

#if !defined(PURE_OSS)
	    repeat.enable= 1;
	    if (ioctl(audout_fd[chanID], HAL_BRCM_IOCTL_AUDIO_SET_AUTOREPEAT, &repeat) == -1)
	    {
		TRACE_IO("IOCTL_AUDIO_SET_AUTOREPEAT")
		return -1;
	    }
#endif

#if !defined(PURE_OSS)
	    /* Set 'rate.ratio' to something sane compared to the set of
	     * SNDCTL_DSP_SPEED settings we'll use to generate the
	     * different tones. This way round appears to allow us a) to
	     * keep the sample we send small, and b) only to send the
	     * sample to the hardware once.
	     */
	    rate.ratio= HAL_OUTPUT_SAMPLE_RATE;
	    if (ioctl(audout_fd[chanID], HAL_BRCM_IOCTL_SET_SAMPLE_RATE, &rate) == -1)
	    {
		TRACE_IO("IOCTL_SET_SAMPLE_RATE")
		return -1;
	    }
#endif

	    /* NB: write() traditionally returns (ssize_t)n > 0 [successful
	     * output quantity], but broadcom PCM hardware doesn't appear
	     * to do this *when 'autorepeat' is TRUE*
	     */
	    if (write(audout_fd[chanID], &sample_block, HAL_TONE_SAMPLE_MAX) != 0 )
	    {
		return -1;
	    }

	    if (ioctl(audout_fd[chanID], SNDCTL_DSP_SPEED, &speed) == -1)
	    {
		/* Change SNDCTL_DSP_SPEED (traditionally the output sample
		 * quality ... but see comment about only sending the sample to
		 * the hardware once).
		 */
		return -1;
	    }

	    ix= 1;
	    while (ix == 1)
	    {
	      struct timeval	timeout;
	      fd_set		read_fds;
	      int		select_n;
	      static int	idx= 0;

		FD_ZERO(&read_fds);
		FD_SET(fileno(stdin), &read_fds);
		timeout.tv_sec= 0;
		timeout.tv_usec= 250000;

		if ((select_n= select(fileno(stdin)+1,
			    &read_fds, NULL, NULL, &timeout)) == -1)
		{
		    perror("select()");
		    exit(1);
		}
		else if (select_n > 0)	/* FDs ready */
		{
		    ix= 0;	/* quit loop */
		}

		printf("%c\b", "|/-\\"[idx++]);
		fflush(stdout);
		idx= idx % 4;
	    }

	    keyb_halt(&sanitty);

	  return 0;
	}
}

int audout_close(const int chanID)
{
	if (chanID >= HAL_MAX_PLAY_OUTPUTS)
	{
	    return -1;
	}
	else if (audout_fd[chanID] == -1)
	{
	    return -1;
	}
	else if (close(audout_fd[chanID]) == -1)
	{
	    return -1;
	}

	audout_fd[chanID]= -1;
  return 0;
}

int process_sample(int chanID, int speed, FILE *samfh)
{
	if (samfh == NULL)
	{
	    TRACE_IO("Bad samfh")
	}
	else if (chanID >= HAL_MAX_PLAY_OUTPUTS)
	{
	    TRACE_IO("Bad chanID")
	    return -1;
	}
	else if (audout_fd[chanID] == -1)
	{
	    TRACE_IO("Channel not open")
	    return -1;	/* already open */
	}
	else
	{
	  int			argument;
	  fd_set		write_fds;
	  u_int8_t		samp_ub[BYTES_PER_KB];

	    argument= speed;
	    if (ioctl(audout_fd[chanID], SNDCTL_DSP_SPEED, &argument) == -1)
	    {
		TRACE_IO("SNDCTL_DSP_SPEED")
		return -1;
	    }
	    if (argument != speed)
	    {
		TRACE_IO("SNDCTL_DSP_SPEED: request not supported")
		return -1;
	    }

#if 0
/* OSS v4.0+ SNDCTL_DSP_GETBLKSIZE is obsolete */
	    /* [=12850] requires rate, chanID, and format set */
	    if (ioctl(audout_fd[chanID], SNDCTL_DSP_GETBLKSIZE, &argument) == -1)
	    {
		TRACE_IO("SNDCTL_DSP_GETBLKSIZE")
		return -1;
	    }
	    else
	    {
		printf("SNDCTL_DSP_GETBLKSIZE: %d\n", argument);
	    }
#endif

#if 0 /* fails on broadcom hardware */
	    /* SNDCTL_DSP_GETODELAY: bytes until next write played */
	    if (ioctl(audout_fd[chanID], SNDCTL_DSP_GETODELAY, &argument) == -1)
	    {
		TRACE_IO("SNDCTL_DSP_GETODELAY")
		return -1;
	    }
	    else
	    {
		printf("SNDCTL_DSP_GETODELAY: %d\n", argument);
	    }
#endif
#if 0 /* fails on broadcom hardware */
{
count_info ci;
	    /* SNDCTL_DSP_GETOPTR: bytes until next write played */
	    if (ioctl(audout_fd[chanID], SNDCTL_DSP_GETOPTR, &ci) == -1)
	    {
		TRACE_IO("SNDCTL_DSP_GETOPTR")
		printf("ioctl() fail -- SNDCTL_DSP_GETOPTR\n");
		return -1;
	    }

		printf("SNDCTL_DSP_GETOPTR: %d\n", ci.bytes);
}
#endif

	    while ( !feof(samfh) /* audio pending */ )
	    {
	      size_t		more;

		/* populate incoming buffer */
		more= fread(&samp_ub, sizeof(u_int8_t), BYTES_PER_KB, samfh);
		if (more == 0)
		{
		    TRACE_IO("fread()")
		    return -1;
		}

		/* ensure playing */
		if (write(audout_fd[chanID], &samp_ub, more) < more)
		{
		    TRACE_IO("write()")
		    return -1;
		}
	    }


	    /* wait for FD-ready, in case write() doesn't */
	    FD_ZERO(&write_fds);
	    FD_SET(audout_fd[chanID], &write_fds);
	    select(audout_fd[chanID]+1, NULL, &write_fds, NULL, NULL);

	  return 0;
	}
}

int main(const int argc, const char **argv)
{
  int		chanID= 0;
  int		freq;
  FILE	 	*fh= NULL;

	if (argc == 1)
	{
	    freq= 4*1300;	/* 1300 is HAL_TONE_BASE */
	}
	else
	{
	    freq= atoi(argv[1]);
	    if (freq < 100)
	    {
		fprintf(stderr, "%s: Freq-arg %d is ridiculous\n", argv[0], freq);
		exit(1);
	    }

	    if (argc == 3)
	    {
		fh= fopen(argv[2], "rb");
		if (fh == NULL)
		{
		    fprintf(stderr, "%s: Can't open file: %s\n", argv[0], argv[2]);
		    exit(1);
		}
	    }
	}

	printf("Open world %d\n", audout_open(chanID));
	printf("Init world %d\n", audout_config(chanID));
	if (fh != NULL)
	{
	    printf("File world %d\n", process_sample(0, freq, fh));
	}
	else
	{
	    printf("Sine world %d\n", audout_sine(chanID, freq));
	}
	printf("Close world %d\n", audout_close(chanID));

	if (fh != NULL)
	{
	    fclose(fh);
	}
 return 0;
}
